from mock import Mock
from remoto import process


class TestExtendPath(object):

    def setup(self):
        self.path = '/usr/local/bin:/bin:/usr/bin:/usr/local/sbin:/usr/sbin:/sbin'

    def test_no_environment_sets_path(self):
        fake_conn = Mock()
        fake_conn.gateway.remote_exec.return_value = fake_conn
        fake_conn.receive.return_value = {}
        result = process.extend_env(fake_conn, {})
        assert result['env']['PATH'] == self.path

    def test_custom_path_does_not_get_overridden(self):
        fake_conn = Mock()
        fake_conn.gateway.remote_exec.return_value = fake_conn
        fake_conn.receive.return_value = {'PATH': '/home/alfredo/bin'}
        result = process.extend_env(fake_conn, {})
        new_path = result['env']['PATH']
        assert new_path.endswith(self.path)
        assert '/home/alfredo/bin' in new_path

    def test_custom_env_var_extends_existing_env(self):
        fake_conn = Mock()
        fake_conn.gateway.remote_exec.return_value = fake_conn
        fake_conn.receive.return_value = {'PATH': '/home/alfredo/bin'}
        result = process.extend_env(fake_conn, {'extend_env': {'CEPH_VOLUME_DEBUG': '1'}})
        new_path = result['env']['PATH']
        assert result['env']['PATH'].endswith(self.path)
        assert result['env']['CEPH_VOLUME_DEBUG'] == '1'

    def test_extend_env_gets_removed(self):
        fake_conn = Mock()
        fake_conn.gateway.remote_exec.return_value = fake_conn
        fake_conn.receive.return_value = {'PATH': '/home/alfredo/bin'}
        result = process.extend_env(fake_conn, {'extend_env': {'CEPH_VOLUME_DEBUG': '1'}})
        assert result.get('extend_env') is None



from . import BaseConnection
import socket


class LocalConnection(BaseConnection):

    def __init__(self, **kw):
        # hostname gets ignored, and forced to be localhost always
        kw.pop('hostname', None)
        super(LocalConnection, self).__init__(
            hostname='localhost',
            detect_sudo=False,
            **kw
        )

    def _make_connection_string(self, hostname, _needs_ssh=None, use_sudo=None):
        interpreter = self.interpreter
        if use_sudo is not None:
            if use_sudo:
                interpreter = 'sudo ' + interpreter
        elif self.sudo:
            interpreter = 'sudo ' + interpreter
        return 'popen//python=%s' % interpreter



from . import BaseConnection


class KubernetesConnection(BaseConnection):

    executable = 'kubectl'
    remote_import_system = 'json'

    def __init__(self, pod_name, namespace=None, **kw):
        self.namespace = namespace
        self.pod_name = pod_name
        super(KubernetesConnection, self).__init__(hostname='localhost', **kw)

    def command_template(self):
        if self.namespace:
            prefix = [
                self.executable, 'exec', '-i', '-n',
                self.namespace, self.pod_name, '--', '/bin/sh', '-c'
            ]
        else:
            prefix = [
                self.executable, 'exec', '-i',
                self.pod_name, '--', '/bin/sh', '-c'
            ]
        return prefix

    def cmd(self, cmd):
        tmpl = self.command_template()
        tmpl.append(' '.join(cmd))
        return tmpl



from pytest import raises
from remoto import log
from remoto.exc import TimeoutError
from mock import Mock


class TestReporting(object):

    def test_reporting_when_channel_is_empty(self):
        conn = Mock()
        result = Mock()
        result.receive.side_effect = EOFError
        log.reporting(conn, result)

    def test_write_debug_statements(self):
        conn = Mock()
        result = Mock()
        result.receive.side_effect = [{'debug': 'a debug message'}, EOFError]
        log.reporting(conn, result)
        assert conn.logger.debug.called is True
        assert conn.logger.info.called is False

    def test_write_info_statements(self):
        conn = Mock()
        result = Mock()
        result.receive.side_effect = [{'error': 'an error message'}, EOFError]
        log.reporting(conn, result)
        assert conn.logger.debug.called is False
        assert conn.logger.error.called is True

    def test_strip_new_lines(self):
        conn = Mock()
        result = Mock()
        result.receive.side_effect = [{'error': 'an error message\n\n'}, EOFError]
        log.reporting(conn, result)
        message = conn.logger.error.call_args[0][0]
        assert message == 'an error message'

    def test_strip_new_line(self):
        conn = Mock()
        result = Mock()
        result.receive.side_effect = [{'error': 'an error message\n'}, EOFError]
        log.reporting(conn, result)
        message = conn.logger.error.call_args[0][0]
        assert message == 'an error message'

    def test_strip_new_line_and_carriage_return(self):
        conn = Mock()
        result = Mock()
        result.receive.side_effect = [{'error': 'an error message\r\n'}, EOFError]
        log.reporting(conn, result)
        message = conn.logger.error.call_args[0][0]
        assert message == 'an error message'

    def test_strip_return(self):
        conn = Mock()
        result = Mock()
        result.receive.side_effect = [{'error': 'an error message\r'}, EOFError]
        log.reporting(conn, result)
        message = conn.logger.error.call_args[0][0]
        assert message == 'an error message'

    def test_timeout_error(self):
        conn = Mock()
        result = Mock()
        result.receive.side_effect = TimeoutError
        log.reporting(conn, result)
        message = conn.logger.warning.call_args[0][0]
        assert 'No data was received after ' in message

    def test_raises_other_errors(self):
        conn = Mock()
        result = Mock()
        result.receive.side_effect = OSError
        with raises(OSError):
            log.reporting(conn, result)



import pytest
from remoto.connection import get


base_names = [
    'ssh', 'oc', 'openshift', 'kubernetes', 'k8s', 'local', 'popen', 'localhost', 'docker', 'podman',
]

capitalized_names = [n.capitalize() for n in base_names]

spaced_names = [" %s " % n for n in base_names]

valid_names = base_names + capitalized_names + spaced_names


class TestGet(object):

    @pytest.mark.parametrize('name', valid_names)
    def test_valid_names(self, name):
        conn_class = get(name)
        assert conn_class.__name__.endswith('Connection')

    def test_fallback(self):
        conn_class = get('non-existent')
        assert conn_class.__name__ == 'BaseConnection'

    def test_custom_fallback(self):
        conn_class = get('non-existent', 'openshift')
        assert conn_class.__name__ == 'OpenshiftConnection'



from . import BaseConnection


class DockerConnection(BaseConnection):
    """
    This connection class allows to (optionally) define a remote hostname
    to connect that holds a given container::

        >>> conn = DockerConnection(hostname='srv-1', container_id='asdf-lkjh')

    Either ``container_id`` or ``container_name`` can be provided to connect to
    a given container.

    .. note:: ``hostname`` defaults to 'localhost' when undefined
    """

    executable = 'docker'
    remote_import_system = 'json'

    def __init__(self, hostname=None, container_id=None, container_name=None, user=None, **kw):
        self.hostname = hostname or 'localhost'
        self.identifier = container_id or container_name
        if not self.identifier:
            raise TypeError('Either container_id or container_name must be provided')
        self.user = user
        super(DockerConnection, self).__init__(hostname=self.hostname, **kw)

    def command_template(self):
        if self.user:
            prefix = [
                self.executable, 'exec', '-i',
                '-u', self.user,
                self.identifier, '/bin/sh', '-c'
            ]
        else:
            prefix = [
                self.executable, 'exec', '-i',
                self.identifier, '/bin/sh', '-c'
            ]
        return prefix

    def cmd(self, cmd):
        tmpl = self.command_template()
        tmpl.append(' '.join(cmd))
        return tmpl



"""
this is just a stub module to use to test the `import_module` functionality in
remoto
"""
import sys


def function(conn):
    return True


def fails():
    raise Exception('failure from fails() function')


def unexpected_fail():
    sys.exit(1)


def noop():
    sys.exit(0)


def passes():
    pass



import logging
# compatibility for older clients that rely on the previous ``Connection`` class
from remoto.backends import BaseConnection as Connection # noqa
from remoto.backends import ssh, openshift, kubernetes, local, podman, docker


logger = logging.getLogger('remoto')


def get(name, fallback='ssh'):
    """
    Retrieve the matching backend class from a string. If no backend can be
    matched, it raises an error.

    >>> get('ssh')
    <class 'remoto.backends.BaseConnection'>
    >>> get()
    <class 'remoto.backends.BaseConnection'>
    >>> get('non-existent')
    <class 'remoto.backends.BaseConnection'>
    >>> get('non-existent', 'openshift')
    <class 'remoto.backends.openshift.OpenshiftConnection'>
    """
    mapping = {
        'ssh': ssh.SshConnection,
        'oc': openshift.OpenshiftConnection,
        'openshift': openshift.OpenshiftConnection,
        'kubernetes': kubernetes.KubernetesConnection,
        'k8s': kubernetes.KubernetesConnection,
        'local': local.LocalConnection,
        'popen': local.LocalConnection,
        'localhost': local.LocalConnection,
        'docker': docker.DockerConnection,
        'podman': podman.PodmanConnection,
    }
    if not name:
        # fallsback to just plain local/ssh
        name = 'ssh'

    name = name.strip().lower()
    connection_class = mapping.get(name)
    if not connection_class:
        logger.warning('no connection backend found for: "%s"' % name)
        if fallback:
            logger.info('falling back to "%s"' % fallback)
            # this assumes that ``fallback`` is a valid mapping name
            return mapping.get(fallback)
    return connection_class



from remoto import util


class TestAdminCommand(object):

    def test_prepend_list_if_sudo(self):
        result = util.admin_command(True, ['ls'])
        assert result == ['sudo', 'ls']

    def test_skip_prepend_if_not_sudo(self):
        result = util.admin_command(False, ['ls'])
        assert result == ['ls']

    def test_command_that_is_not_a_list(self):
        result = util.admin_command(True, 'ls')
        assert result == ['sudo', 'ls']


class TestRemoteError(object):

    def setup(self):
        self.traceback = ('\n').join([
            'Traceback (most recent call last):',
            '  File "<string>", line 1, in <module>',
            "NameError: name 'foo' is not defined"
        ])

    def test_exception_name(self):
        error = util.RemoteError(self.traceback)
        assert error.exception_name == 'NameError'

    def test_exception_line(self):
        error = util.RemoteError(self.traceback)
        assert error.exception_line == "NameError: name 'foo' is not defined"



import pytest


class Capture(object):

    def __init__(self, *a, **kw):
        self.a = a
        self.kw = kw
        self.calls = []
        self.return_values = kw.get('return_values', False)
        self.always_returns = kw.get('always_returns', False)

    def __call__(self, *a, **kw):
        self.calls.append({'args': a, 'kwargs': kw})
        if self.always_returns:
            return self.always_returns
        if self.return_values:
            return self.return_values.pop()


class Factory(object):

    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)



import inspect
import json
import socket
import sys
import execnet
import logging
from remoto.process import check


class BaseConnection(object):
    """
    Base class for Connection objects. Provides a generic interface to execnet
    for setting up the connection
    """
    executable = ''
    remote_import_system = 'legacy'

    def __init__(self, hostname, logger=None, sudo=False, threads=1, eager=True,
                 detect_sudo=False, interpreter=None, ssh_options=None):
        self.sudo = sudo
        self.hostname = hostname
        self.ssh_options = ssh_options
        self.logger = logger or basic_remote_logger()
        self.remote_module = None
        self.channel = None
        self.global_timeout = None  # wait for ever

        self.interpreter = interpreter or 'python%s' % sys.version_info[0]

        if eager:
            try:
                if detect_sudo:
                    self.sudo = self._detect_sudo()
                self.gateway = self._make_gateway(hostname)
            except OSError:
                self.logger.error(
                    "Can't communicate with remote host, possibly because "
                    "%s is not installed there" % self.interpreter
                )
                raise

    def _make_gateway(self, hostname):
        gateway = execnet.makegateway(
            self._make_connection_string(hostname)
        )
        gateway.reconfigure(py2str_as_py3str=False, py3str_as_py2str=False)
        return gateway

    def _detect_sudo(self, _execnet=None):
        """
        ``sudo`` detection has to create a different connection to the remote
        host so that we can reliably ensure that ``getuser()`` will return the
        right information.

        After getting the user info it closes the connection and returns
        a boolean
        """
        exc = _execnet or execnet
        gw = exc.makegateway(
            self._make_connection_string(self.hostname, use_sudo=False)
        )

        channel = gw.remote_exec(
            'import getpass; channel.send(getpass.getuser())'
        )

        result = channel.receive()
        gw.exit()

        if result == 'root':
            return False
        self.logger.debug('connection detected need for sudo')
        return True

    def _make_connection_string(self, hostname, _needs_ssh=None, use_sudo=None):
        _needs_ssh = _needs_ssh or needs_ssh
        interpreter = self.interpreter
        if use_sudo is not None:
            if use_sudo:
                interpreter = 'sudo ' + interpreter
        elif self.sudo:
            interpreter = 'sudo ' + interpreter
        if _needs_ssh(hostname):
            if self.ssh_options:
                return 'ssh=%s %s//python=%s' % (
                    self.ssh_options, hostname, interpreter
                )
            else:
                return 'ssh=%s//python=%s' % (hostname, interpreter)
        return 'popen//python=%s' % interpreter

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.exit()
        return False

    def cmd(self, cmd):
        """
        In the base connection class, this method just returns the ``cmd``
        as-is. Other implementations will end up doing transformations to the
        command by prefixing it with other flags needed. See
        :class:`KubernetesConnection` for an example
        """
        return cmd

    def execute(self, function, **kw):
        return self.gateway.remote_exec(function, **kw)

    def exit(self):
        self.gateway.exit()

    def import_module(self, module):
        """
        Allows remote execution of a local module. Depending on the
        ``remote_import_system`` attribute it may use execnet's implementation
        or remoto's own based on JSON.

        .. note:: It is not possible to use execnet's remote execution model on
                  connections that aren't SSH or Local.
        """
        if self.remote_import_system is not None:
            if self.remote_import_system == 'json':
                self.remote_module = JsonModuleExecute(self, module, self.logger)
            else:
                self.remote_module = LegacyModuleExecute(self.gateway, module, self.logger)
        else:
            self.remote_module = LegacyModuleExecute(self.gateway, module, self.logger)
        return self.remote_module


class LegacyModuleExecute(object):
    """
    This (now legacy) class, is the way ``execnet`` does its remote module
    execution: it sends it over a channel, and does a send/receive for
    exchanging information. This only works when there is native support in
    execnet for a given connection. This currently means it would only work for
    ssh and local (Popen) connections, and will not work for anything like
    kubernetes or containers.
    """

    def __init__(self, gateway, module, logger=None):
        self.channel = gateway.remote_exec(module)
        self.module = module
        self.logger = logger

    def __getattr__(self, name):
        if not hasattr(self.module, name):
            msg = "module %s does not have attribute %s" % (str(self.module), name)
            raise AttributeError(msg)
        docstring = self._get_func_doc(getattr(self.module, name))

        def wrapper(*args):
            arguments = self._convert_args(args)
            if docstring:
                self.logger.debug(docstring)
            self.channel.send("%s(%s)" % (name, arguments))
            try:
                return self.channel.receive()
            except Exception as error:
                # Error will come as a string of a traceback, remove everything
                # up to the actual exception since we do get garbage otherwise
                # that points to non-existent lines in the compiled code
                exc_line = str(error)
                for tb_line in reversed(str(error).split('\n')):
                    if tb_line:
                        exc_line = tb_line
                        break
                raise RuntimeError(exc_line)

        return wrapper

    def _get_func_doc(self, func):
        try:
            return getattr(func, 'func_doc').strip()
        except AttributeError:
            return ''

    def _convert_args(self, args):
        if args:
            if len(args) > 1:
                arguments = str(args).rstrip(')').lstrip('(')
            else:
                arguments = str(args).rstrip(',)').lstrip('(')
        else:
            arguments = ''
        return arguments


dump_template = """
if __name__ == '__main__':
    import json, traceback
    obj = {'return': None, 'exception': None}
    try:
        obj['return'] = %s%s
    except Exception:
        obj['exception'] = traceback.format_exc()
    try:
        print(json.dumps(obj).decode('utf-8'))
    except AttributeError:
        print(json.dumps(obj))
"""


class JsonModuleExecute(object):
    """
    This remote execution class allows to ship Python code over to the remote
    node, load it via ``stdin`` and call any function with arguments. The
    resulting response is dumped over JSON so that it can get printed to
    ``stdout``, then captured locally, loaded into regular Python and returned.

    If the remote end generates an exception with a traceback, that is captured
    as well and raised accordingly.
    """

    def __init__(self, conn, module, logger=None):
        self.conn = conn
        self.module = module
        self._module_source = inspect.getsource(module)
        self.logger = logger

    def __getattr__(self, name):
        if not hasattr(self.module, name):
            msg = "module %s does not have attribute %s" % (str(self.module), name)
            raise AttributeError(msg)
        docstring = self._get_func_doc(getattr(self.module, name))

        def wrapper(*args):
            if docstring:
                self.logger.debug(docstring)
            if len(args):
                source = self._module_source + dump_template % (name, repr(args))
            else:
                source = self._module_source + dump_template % (name, '()')

            out, err, code = check(self.conn, ['python'], stdin=source.encode('utf-8'))
            if not out:
                if not err:
                    err = [
                        'Traceback (most recent call last):',
                        '    File "<stdin>", in <module>',
                        'Exception: error calling "%s"' % name
                    ]
                if code:
                    raise Exception('Unexpected remote exception: \n%s' % '\n'.join(err))
                # at this point, there was no stdout, and the exit code was 0,
                # we must return so that we don't fail trying to serialize back
                # the JSON
                return
            response = json.loads(out[0])
            if response['exception']:
                raise Exception(response['exception'])
            return response['return']

        return wrapper

    def _get_func_doc(self, func):
        try:
            return getattr(func, 'func_doc').strip()
        except AttributeError:
            return ''


def basic_remote_logger():
    logging.basicConfig()
    logger = logging.getLogger(socket.gethostname())
    logger.setLevel(logging.DEBUG)
    return logger


def needs_ssh(hostname, _socket=None):
    """
    Obtains remote hostname of the socket and cuts off the domain part
    of its FQDN.
    """
    if hostname.lower() in ['localhost', '127.0.0.1', '127.0.1.1']:
        return False
    _socket = _socket or socket
    fqdn = _socket.getfqdn()
    if hostname == fqdn:
        return False
    local_hostname = _socket.gethostname()
    local_short_hostname = local_hostname.split('.')[0]
    if local_hostname == hostname or local_short_hostname == hostname:
        return False
    return True



from mock import Mock, patch
from remoto import file_sync


class TestRsync(object):

    def make_fake_sync(self):
        fake_sync = Mock()
        fake_sync.return_value = fake_sync
        fake_sync.targets = []
        fake_sync.add_target = lambda gw, destination: fake_sync.targets.append(destination)
        return fake_sync

    @patch('remoto.file_sync.Connection', Mock())
    def test_rsync_fallback_to_host_list(self):
        fake_sync = self.make_fake_sync()
        with patch('remoto.file_sync._RSync', fake_sync):
            file_sync.rsync('host1', '/source', '/destination')

        # should've added just one target
        assert len(fake_sync.targets) == 1

    @patch('remoto.file_sync.Connection', Mock())
    def test_rsync_use_host_list(self):
        fake_sync = self.make_fake_sync()
        with patch('remoto.file_sync._RSync', fake_sync):
            file_sync.rsync(
                ['host1', 'host2'], '/source', '/destination')

        # should've added just one target
        assert len(fake_sync.targets) == 2
